# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os

from pytorch_lightning import Trainer

from .callbacks import PredictionsWriter

logger = logging.getLogger(__name__)


def run(model, data_loader, name, device, output_folder, grad_cams=False):
    """Runs inference on input data, outputs HDF5 files with predictions.

    Parameters
    ---------
    model : :py:class:`torch.nn.Module`
        neural network model (e.g. pasa)

    data_loader : py:class:`torch.torch.utils.data.DataLoader`

    name : str
        the local name of this dataset (e.g. ``train``, or ``test``), to be
        used when saving measures files.

    device : str
        device to use ``cpu`` or ``cuda:0``

    output_folder : str
        folder where to store output prediction and model
        summary

    grad_cams : bool
        if we export grad cams for every prediction (must be used along
        a batch size of 1 with the DensenetRS model)

    Returns
    -------

    all_predictions : list
        All the predictions associated with filename and groundtruth
    """
    output_folder = os.path.join(output_folder, name)

    logger.info(f"Output folder: {output_folder}")
    os.makedirs(output_folder, exist_ok=True)

    logger.info(f"Device: {device}")

    logfile_name = os.path.join(output_folder, "predictions.csv")
    logfile_fields = ("filename", "likelihood", "ground_truth")

    trainer = Trainer(
        accelerator="auto",
        devices="auto",
        callbacks=[
            PredictionsWriter(
                logfile_name=logfile_name,
                logfile_fields=logfile_fields,
                write_interval="epoch",
            ),
        ],
    )

    all_predictions = trainer.predict(model, data_loader)

    return all_predictions
